ggamm<-function(dep,
	family="normal",
	link="logit",
	fix=matrix("NA",1,1),
	smooth=matrix("NA",1,1),
	nknot=rep(20,dim(as.matrix(smooth))[2]),
	ord=rep(2,dim(as.matrix(smooth))[2]),
	deg=rep(3,dim(as.matrix(smooth))[2]),
	plotf=F,
        include.lin=F,
	reg=matrix("NA",1,1),
        regions=reg,
	pmatrix=matrix(0,1,1),
	map,
	plotmap=F,
	random=matrix("NA",1,1),
	id=matrix("NA",1,1),
	weight=rep(1,length(dep)),
        offset=rep(0,length(dep)),
	sig=0.95,
	eps=0.00001,
	noprint=F,
	nowarnings=F,
        startValue1=0.1,
        startValue2=0.5,
	maxit=500,
	ranktest=F,
	diagmult=1.0005,
	dispers=F,
	outfile="NA",
	log.like=F,
	lowerlim=0.001,
        method="REML"
	)
{
#	ggamm - a software for empirical Bayes inferenc in generalised geoadditive mixed models
#	version 0.2
#	author: 	Thomas Kneib
#       email:		kneib@stat.uni-muenchen.de
#	last modified:	12.08.2003

#	dep		the dependent variable
#	family		distributional assumption for y. one out of "normal", "binomial", "poisson" and "gamma"
#	link		the link-function. one out of "logit" and "probit" (only meaningful if family==binomial)
#			for the other distributions the response functions identity (normal) or exp (poisson, gamma) are used
#	dispers		whether to estimate dispersion-parameter for binomial or poisson distributed response
#	fix		matrix containing fixed effects covariates
#	smooth		matrix containing smooth covariates (modelled as p-splines)
#	nknot		vector containing the number of inner knots for every smooth covariate
#	ord		vector conatining the order of the difference penalty for every smooth covariate
#	deg		vector conatining the degree of the b-spline basis for every smooth covariate
#	plotf		whether to plot the estimated functions or not
#	include.lin	whether to include the linear part of the functions into the plot or not
#	reg		vector containing the spatial covariate
#	pmatrix		penalty matrix corresponding to the neighbourhood-structure of the spatial covariate
#       regions		vector containing the distinct values of the spatial covariate (in the order corresponding to the penalty matrix )
#	map		map-object, used to plot the spatial effect (created with the function readbndfile, compare Brezger et. al. 2003)
#	plotmap		whether to plot the estimated spatial effect or not
#	random		matrix containing the realisations of the covariates that should be modelled as random effects. a random intercept must be specified as a column of '1's
#	id		vector containing the idetification variable of the clusters for the random effects
#	weight		weight-variable
#       offset		variable containing the offset
#	sig		nominal level of the computed (pointwise) confidence intervals
#	eps		Termination condition for the estimation of regression and variance parameters
#	noprint		whether to print informations on the iterations and the estimators or not
#	startValue1	starting-value for the variance/inverse smoothing parameters (of the p-splines and the markov random fields)
#	startValue2	starting-value for the variances of the random effects
#	maxit		maximum number of iterations
#	ranktest	whether to test for rank-deficiency of the fisher-information-matrix or not
#	outfile		if outfile is specified, results are written to the specified directory
#			e.g. if outfile="c:\\temp\\res" estimates of the fixed effects are written to c:\\temp\\res_fixedEffects.raw
#	log.like	whether to compute the (restricted) log-likelihood or not (only meaningful if family=="normal")
#	lowerlim	if the penalized part of a covariate relative to the total linear predictor is smaller then lowerlim (measured in euclidean norm),
#			the estimation of the corresponding variance component is stopped (to avoid numerical problems)
#	method		method for the estimation of the variance components. one out of "REML" and "ML" with "REML" being the default

# convert input into matrices
dep<-as.matrix(dep)
fix<-as.matrix(fix)
smooth<-as.matrix(smooth)
reg<-as.matrix(reg)
random<-as.matrix(random)
id<-as.matrix(id)

cilstring<-paste("ci0p",substring(as.character((1-sig)/2),3,10),sep="")
ciustring<-paste("ci0p",substring(as.character(1-(1-sig)/2),3,10),sep="")
ciblstring<-paste("cib0p",substring(as.character((1-sig)/2),3,10),sep="")
cibustring<-paste("cib0p",substring(as.character(1-(1-sig)/2),3,10),sep="")

# older definition of sig
sig<-(1-sig)/2
lrtest<-F
ki<-2

# substract offset (normal case only)
if(family=="normal")
  {
  y<-dep-offset
  }

# check for unknown family-names
if(family!="normal" && family!="poisson" && family!="binomial" && family!="gamma")
  {
  cat(paste("ERROR: ",family," is not a valid family-name.\n",sep=""))
  cat("       family should be one out of \"normal\", \"binomial\", \"poisson\" or \"gamma\".\n")
  return(invisible())
  }

# number of observations
n<-dim(dep)[1]

# check whether there is a grouping variable for the random effects
if(random[1,1]!="NA" && id[1,1]=="NA")
  {
  cat("ERROR: random effects specified without grouping variable. program stopped.\n")
  return(invisible())
  }

# check whether there is a penalty-matrix for spatial effects
if(reg[1,1]!="NA")
  {
  if(dim(pmatrix)[1]==1)
    {
    cat("ERROR: no penalty matrix for spatial effect. program stopped.\n")
    return(invisible())
    }
  }

# change dispers to T if family is normal or gamma
if((family=="normal"||family=="gamma"))
  {
  if(dispers==F)
    {
    if(nowarnings==F)
      {
      cat("WARNING: dispers==F can only be used with poisson or binomial distributed data. dispers is changed to T.\n\n")
      }
    dispers<-T
    }
  }

# in the absence of smooth, spatial and random covariates call the splus-function glm
if((smooth[1,1]=="NA" && reg[1,1]=="NA") && (random[1,1]=="NA"))
  {

# no offset specification possible in function glm
  if(sum(abs(offset-rep(0,n)))!=0)
    {
    cat("ERROR: program does not support usage of offset in absence of smooth, spatial and random effects. program stopped.\n")
    return(invisible())
    }

# define the formula of the glm
  if(fix[1,1]=="NA")
    {
    glm1.form<-y~1
    q1<-0
    }
  else
    {
    glm1.form<-y~fix
    q1<-dim(fix)[2]
    }

#call to glm with appropriate choice of family and link
  if(family=="normal")
    {
    glm1<-glm(glm1.form,family=gaussian,trace=(!noprint),weights=weight,x=T)
    }
  else if(family=="gamma")
    {
    y<-as.vector(dep)
    glm1<-glm(glm1.form,family=Gamma(link="log"),trace=(!noprint),weights=weight,x=T)
    }
  else if(family=="poisson")
    {
    y<-as.vector(dep)
    if(dispers==F)
      {
      glm1<-glm(glm1.form,family=poisson,trace=(!noprint),weights=weight,x=T)
      }
    else
      {
      glm1<-glm(glm1.form,family=quasi(link=log,variance=mu),trace=(!noprint),weights=weight,x=T)
      }
    }
  else if(family=="binomial")
    {
    y<-as.vector(dep)
    if(dispers==F)
      {
      if(link=="logit")
        {
        glm1<-glm(glm1.form,family=binomial,trace=(!noprint),weights=weight,x=T)
        }
      else if(link=="probit")
        {
        glm1<-glm(glm1.form,family=binomial(link="probit"),trace=(!noprint),weights=weight,x=T)
        }
      }
    else
      {
      if(link=="logit")
        {
        glm1<-glm(glm1.form,family=quasi(link="logit",variance="mu(1-mu)"),trace=(!noprint),weights=weight,x=T)
        }
      else if(link=="probit")
        {
        glm1<-glm(glm1.form,family=quasi(link="probit",variance="mu(1-mu)"),trace=(!noprint),weights=weight,x=T)
        }
      }
    }

  beta<-glm1$coef

#get the number of iterations
  iterations<-glm1$iter

# compute the dispersion paramater
  if(dispers==T)
    {
    if(method=="REML")
      {
      theta<-as.matrix(sum((glm1$residuals)^2*glm1$weights)/(n-dim(fix)[2]))
      }
    else
      {
      theta<-as.matrix(sum((glm1$residuals)^2*glm1$weights)/n)
      }
    dimnames(theta)<-list(NULL,"phi")
    if(noprint==F)
      {
      cat(paste("phi: ",round(theta[1,1],5),"\n",sep=""))
      }
    }

#get the fixed effects and compute confidence intervals
  names1<-rep(" ",q1+1)
  for(j in 1:(q1+1))
    {
    if(noprint==F)
      {
      cat(paste(paste("beta",(j-1),sep=""),paste(":",round(glm1$coef[j],5)),"\n",sep=""))
      }
    names1[j]<-paste("beta",(j-1),sep="")
    }
  names2<-c("coefficient","std",cilstring,ciustring)
  fixedEffect<-matrix(0,q1+1,4)
  fixedEffect[,1]<-glm1$coef
  std<-sqrt(diag(solve(t(glm1$x)%*%diag(glm1$weights)%*%glm1$x)))
  fixedEffect[,2]<-std
  fixedEffect[,3]<-fixedEffect[,1]+qnorm(sig)*std
  fixedEffect[,4]<-fixedEffect[,1]-qnorm(sig)*std
  dimnames(fixedEffect)<-list(names1,names2)

#write estimation results to the specified files
  if(outfile!="NA")
    {
    write.table(round(fixedEffect,5),paste(outfile,"_fixedEffects.raw",sep=""),dimnames.write=T,sep=" ")
    write.table(iterations,paste(outfile,"_iterations.raw",sep=""),dimnames.write="colnames",sep=" ")
    }

#compute log-likelihood and restricted log-likelihood
  if(family=="normal" && log.like==T)
    {
    log.like<-matrix(0,1,1)

#design-matrix
    if(q1>0)
      {
      X<-cbind(rep(1,n),fix)
      }
    else
      {
      X<-as.matrix(rep(1,n))
      }
#residuals
    res<-y-X%*%beta
#vector containing the diagonal elements of V (the covariance matrix of y)
    v<-as.vector(weight*as.double(theta))
    log.like[1,1]<--0.5*(sum(log(v))+multdiagback(t(res),1/v)%*%res)
    if(method=="REML")
      {
      log.like[1,1]<-log.like[1,1]-0.5*logdet(multdiagback(t(X),1/v)%*%X)
      }
    if(noprint==F)
      {
      if(method=="ML")
        {
        cat(paste("log-likelihood: ",round(log.like[1,1],5),"\n",sep=""))
        }
      else
        {
        cat(paste("restricted log-likelihood: ",round(log.like[1,1],5),"\n",sep=""))
        }
      }
    if(outfile!="NA")
      {
      write.table(round(log.like,5),paste(outfile,"_log.like.raw",sep=""),dimnames.write="colnames",sep=" ")
      }
#return results if log.like==T)
    return(list(variances=theta,fixedEffects=fixedEffect,log.like=log.like,iterations=iterations))
    }

#return results
  if(dispers==T)
    {
    write.table(round(theta,5),paste(outfile,"_variances.raw",sep=""),dimnames.write=T,sep=" ")
    return(list(variances=round(theta,5),fixedEffects=round(fixedEffect,5),iterations=iterations))
    }
  else
    {
    return(list(fixedEffects=round(fixedEffect,5),iterations=iterations))
    }
  }

#matrix of the fixed effects (with intercept)
#q1 is the number of fixed effects
if(fix[1,1]=="NA")
  {
  X<-as.matrix(rep(1,n))
  }
else
  {
  X<-cbind(rep(1,n),fix)
  }
q1<-dim(X)[2]


#modify/compute design matrices for the smooth effects
#bcut contains the partition of B corresponding to the different smooth functions
bcut<-1
if(smooth[1,1]!="NA")
  {
#q2 is the number of smooth effects
  q2<-dim(smooth)[2]
  if(length(nknot)!=q2)
    {
    cat("WARNING: length of vector nknot should be the same as the number of variables in smooth.\n")
    cat("         default values substituted.\n\n")
    nknot<-rep(20,q2)
    }
  if(length(ord)!=q2)
    {
    cat("WARNING: length of vector ord should be the same as the number of variables in smooth.\n")
    cat("         default values substituted.\n\n")
    ord<-rep(2,q2)
    }
  if(length(deg)!=q2)
    {
    cat("WARNING: length of vector deg should be the same as the number of variables in smooth.\n")
    cat("         default values substituted.\n\n")
    deg<-rep(3,q2)
    }

#matrix for the random components of the smooth effects
  B<-matrix(0,n,sum(nknot)+sum(deg)-sum(ord)-q2)
  for(j in 1:q2)
    {
#define knots
    xl<-min(smooth[,j])
    xr<-max(smooth[,j])
    xmin<-xl-(xr-xl)/100
    xmax<-xr+(xr-xl)/100
    dx<-(xmax-xmin)/(nknot[j]-1)
    knots<-seq(xmin-deg[j]*dx,xmax+deg[j]*dx,by=dx)
#get the b-spline design-matrix
    bs<-spline.des(knots,smooth[,j],deg[j]+1)
    knots<-knots[1:(nknot[j]+deg[j]-1)]
#difference-matrix of order 2
    D<-matrix(0,nknot[j]+deg[j]-ord[j]-1,nknot[j]+deg[j]-1)                      
    if(ord[j]==2)
      {
      for(i in 1:(nknot[j]+deg[j]-3))
        {
        D[i,i]<-1
        D[i,i+1]<--2
        D[i,i+2]<-1
        }
#parametric part of fj
      X<-cbind(X,center(bs$design%*%knots))
      bcut<-c(bcut,bcut[j]+length(knots)-2)
      }
    else
      {
      for(i in 1:(nknot[j]+deg[j]-2))
        {
        D[i,i]<--1
        D[i,i+1]<-1
        }
      bcut<-c(bcut,bcut[j]+length(knots)-1)
      }
#random part of fj
    B[,bcut[j]:(bcut[j+1]-1)]<-bs$design%*%t(D)%*%solve(D%*%t(D))
    }
  }
else
#no smooth effects
  {
  q2<-0
  }

#compute/modify design matrix for the spatial effect
if(reg[1,1]!="NA")
  {
  Z<-incident(reg,regions)
#eigen-decomposition of the penalty-matrix
  e<-eigen(pmatrix)
#last eigen-value may be negative due to numerical errors
  e$values[length(regions)]<-0
#compute factor L, so that K=LL'
  L<-multdiagback(e$vectors,sqrt(e$values))                             
  L<-L[1:length(regions),1:(length(regions)-1)]
  if(smooth[1,1]!="NA")
    {
    B<-cbind(B,Z%*%multdiagback(L,1/e$values[1:(length(regions)-1)]))
    q2<-q2+1
    }
  else
    {
    B<-Z%*%L%*%solve(t(L)%*%L)
    q2<-1
    }
  bcut<-c(bcut,bcut[length(bcut)]+length(regions)-1)
  }

#compute/modify design matrix for random effects
if(random[1,1]!="NA")
  {
#number of random effects
  q3<-dim(random)[2]
  e<-incident2(id)
#incident matrix corresponding to the grouping structure
  Z<-e$incident
# different values of the grouping-variable
  randoms<-e$randoms

  if((smooth[1,1]=="NA")&&(reg[1,1]=="NA"))
    {
    B<-Z*random[,1]
    }
  else
    {
    B<-cbind(B,Z*random[,1])
    }
  bcut<-c(bcut,bcut[length(bcut)]+length(randoms))
#define design matrix for random slopes as product of Z and the vector of covariates
  if(q3>1)
    {
    for(j in 2:q3)
      {
      B<-cbind(B,Z*random[,j])
      bcut<-c(bcut,bcut[length(bcut)]+length(randoms))
      }
    }
  }
else
# no random effects
  {
  q3<-0
  }

# define list to contain results
if((family=="normal"))
  {
  erg<-vector("list",3+q2+q3+log.like+lrtest)
  names(erg)<-rep(" ",3+q2+q3+log.like+lrtest)
  }
else
  {
  erg<-vector("list",4+q2+q3)
  names(erg)<-rep(" ",4+q2+q3)
  }

#compute starting values for beta (least-squares-estimate)
beta0<-rep(0,dim(X)[2]+dim(B)[2])
if(family=="poisson")
  {
  beta1<-rep(0,dim(X)[2]+dim(B)[2])
  }
else
  {
  if(qr(t(B)%*%B)$rank==dim(B)[2])
    {
    beta1<-c(solve(t(X)%*%X,t(X)%*%dep),solve(t(B)%*%B,t(B)%*%dep))
    }
  else
    {
    beta1<-rep(0,dim(X)[2]+dim(B)[2])
    }
  }

#case 1: family poisson or binomial without dispersion parameter
if((family=="binomial"||family=="poisson")&&(dispers==F))
  {
#starting values for theta
  theta0<-rep(Inf,q2+q3)                                      
	# theta0[1:q2]==inverse smoothing parameters
	# theta0[(q2+1):(q2+q3)]==variances of random effects
  theta1<-c(rep(startValue1,q2),rep(startValue2,q3))
#vector containing the signs of the current estimate of theta
#this vector is needed because of the unrestricted estimation of theta
  signs<-rep(1,q2+q3)

#score-function and expected fisher information for theta
  s<-rep(0,q2+q3)
  Fi<-matrix(0,q2+q3,q2+q3)

#relative changes in beta/theta  
  crit1<-1
  crit2<-1
  stopcrit<-rep(10,q2+q3)
  its<-rep(0,q2+q3)
  test<-T

#number of iteration  
  k<-1

#linear predictor
eta<-offset+X%*%beta1[1:dim(X)[2]]+B%*%beta1[(dim(X)[2]+1):length(beta1)]  

#start estimation loop
  while(test==T)
    {
    if(noprint==F)
      {
      cat(paste("iteration:",k,"\n\n"))
      }
    theta0<-theta1
    beta0<-beta1

#compute weights and y for the working lmm
    if(family=="poisson")
      {
      mu<-exp(eta)
      w<-exp(eta)*weight
      dinv<-exp(-eta)
      }
    else if(family=="binomial")
      {
      if(link=="logit")
        {
        mu<-exp(eta)/(1+exp(eta))
        w<-mu*(1-mu)*weight
        dinv<-1/(mu*(1-mu))
        }
      else if(link=="probit")
        {
        mu<-pnorm(eta)
        w<-weight*dnorm(eta)^2/(mu*(1-mu))
        dinv<-1/dnorm(eta)
        }
      }
    y<-as.matrix(eta-offset+dinv*(dep-mu))

#fisher-scoring step for beta
#if ranktest==T, H is tested for rank-deficiency and if H is singular, the digonal elements are inflated (by multiplication with 1.0005)
    H<-rbind(cbind(t(X)%*%multdiagfront(X,w),t(X)%*%multdiagfront(B,w)),cbind(t(B)%*%multdiagfront(X,w),addtodiag(t(B)%*%multdiagfront(B,w),makeLambdainv(theta0,bcut))))
    H1<-rbind(t(X)%*%(w*y),t(B)%*%(w*y))
    if(ranktest==T)
      {
      rank<-qr(H)$rank
      if(rank<dim(H)[1])
        {
        if(nowarnings==F)
          {
          cat("WARNING: fisher-information was found to be singular.\n")
          cat(paste("         rank is ",rank,", should be ",length(beta1),".\n",sep=""))
          cat(paste("         diagonal elements were multiplied with ",diagmult,".\n\n",sep=""))
          }
        H<-H+diag(diagmult*diag(H))
        }
      }
    beta1<-solve(H,H1)

#compute  new linear predictor
    eta<-X%*%beta1[1:dim(X)[2],]+B%*%beta1[(dim(X)[2]+1):(dim(X)[2]+dim(B)[2]),]
    res<-as.matrix(y-eta)
    eta<-eta+offset

    if(method=="REML")
      {
       Hinv<-solve(H)
       H<-addtodiag(H,-makeLambdainv(theta0,bcut),dim(X)[2]+1,dim(H)[1])

#transform the restricted estimation problem to an unrestricted
    theta0<-signs*sqrt(theta0)

#compute score-function and expected fisher-information for theta
      for(j in 1:(q2+q3))
        {
        s[j]<--0.5*spur(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*theta0[j]*2)+0.5*spur(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*theta0[j]*2)+sum((t(res*w)%*%B[,bcut[j]:(bcut[j+1]-1)])^2)*theta0[j]
        for(k in j:(q2+q3))
          {
          Fi[j,k]<-0.5*spur(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1)]%*%H[(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*4*theta0[j]*theta0[k])-spur((H[(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)])%*%(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1)])*4*theta0[j]*theta0[k])+0.5*spur((H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1)])%*%(H[(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)])*4*theta0[j]*theta0[k])
          Fi[k,j]<-Fi[j,k]
          }
        }
      }

    else if(method=="ML")
      {
      P<-solve(diag(as.vector(1/w))+multdiagback(B,makeLambda(theta1,bcut))%*%t(B))

#transform the restricted estimation problem to an unrestricted
    theta0<-signs*sqrt(theta0)

#compute score-function and expected fisher-information for theta
      for(j in 1:(q2+q3))
        {
        s[j]<--0.5*spur2(P,(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j])%*%t(B[,bcut[j]:(bcut[j+1]-1)]))+0.5*t(res*w)%*%(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j])%*%t(B[,bcut[j]:(bcut[j+1]-1)])%*%(w*res)

        for(k in j:(q2+q3))
          {
          Fi[j,k]<-0.5*spur(P%*%(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j])%*%t(B[,bcut[j]:(bcut[j+1]-1)])%*%P%*%(B[,bcut[k]:(bcut[k+1]-1)]*2*theta0[k])%*%t(B[,bcut[k]:(bcut[k+1]-1)]))
          Fi[k,j]<-Fi[j,k]
          }
        }
      }

#fisher-scoring step for theta
    theta1<-theta0+solve(Fi,s)
    signs<--1*(theta1<0)+1*(theta1>=0)

#transform theta back to the original parametrisation
    theta1<-theta1^2
    theta0<-theta0^2

#stop estimation of theta[i] if stopcrit[i]<lowerlim
    for(i in 1:(q2+q3))
      {
      stopcrit[i]<-sqrt(sum((center(B[,bcut[i]:(bcut[i+1]-1)]%*%beta1[(dim(X)[2]+bcut[i]):(dim(X)[2]+bcut[i+1]-1)]))^2)/sum(center(eta)^2))
      }
    theta1[stopcrit<lowerlim]<-theta0[stopcrit<lowerlim]
    its[stopcrit>lowerlim]<-k

#compute relative changes in beta and theta    
   crit1<-sqrt(sum((beta0-beta1)^2)/sum(beta0^2))
   crit2<-sqrt(sum((theta0-theta1)^2)/sum(theta0^2))

   test<-(((crit1>eps)||(crit2>eps))&&(k<maxit))

   if(noprint==F)
     {
     cat(paste("relative changes in the regression coefficients:",round(crit1,7),"\n"))
     cat(paste("relative changes in the variance parameters:    ",round(crit2,7),"\n\n"))
     }
   k<-k+1
   }

 
 if(k>=maxit)
#algorithm did not converge.
   {
   if(nowarnings==F)
     {
     cat(paste("WARNING: number of iterations reached ",maxit,".\n         algorithm did not converge.\n",sep=""))
     cat("         last two values of tau^2 were:\n")
     print(theta0)
     print(theta1)
     }
   }

  if(reg[1,1]!="NA")
    {
    q2<-q2-1
    }
  if(q2>0)
    {
    for(i in 1:q2)
      {
      if(stopcrit[i]<lowerlim && nowarnings==F)
        {
        cat(paste("WARNING: estimation of the variance of smooth",i," was stopped after iteration ",its[i],"\n         because its penalized part was small relative to the linear predictor.\n\n",sep=""))
        }
      }
    }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    if(stopcrit[q2]<lowerlim && nowarnings==F)
      {
      cat(paste("WARNING: estimation of the variance of the spatial effect was stopped after iteration ",its[q2],"\n         because its penalized part was small relative to the linear predictor.\n\n",sep=""))
      }
    }
  if(q3>0)
    {
    for(i in 1:q3)
      {
      if(stopcrit[q2+i]<lowerlim && nowarnings==F)
        {
        cat(paste("WARNING: estimation of the variance of random",i," was stopped after iteration ",its[q2+i],"\n         because its effect was small relative to the linear predictor.\n\n",sep=""))
        }
      }
    }
  
#extract fix and random effects 
  beta<-beta1[1:dim(X)[2]]
  alpha<-beta1[(dim(X)[2]+1):length(beta1)]

 
#store estimator of theta
  names2<-c("coefficient","stopcrit","stopped")
  names1<-rep(" ",(q2+q3))
  theta0<-matrix(0,q2+q3,3)
  if(reg[1,1]!="NA")
    {
    q2<-q2-1
    }
  if(q2>0)
    {
    for(j in 1:q2)
      {
      if(noprint==F)
        {
        cat(paste(paste("variance of smooth",j,sep=""),paste(":",round(theta1[j],10)),"\n",sep=""))
        }
      names1[j]<-paste("smooth",j,sep="")
      }
    theta0[1:q2,1]<-theta1[1:q2]
    theta0[1:q2,2]<-stopcrit[1:q2]
    theta0[1:q2,3]<-1*(stopcrit[1:q2]<lowerlim)
    }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    if(noprint==F)
      {
      cat(paste(paste("variance of the spatial effect:",round(theta1[q2],10)),"\n",sep=""))
      }
    names1[q2]<-"spatial"
    theta0[q2,1]<-theta1[q2]
    theta0[q2,2]<-stopcrit[q2]
    theta0[q2,3]<-1*(stopcrit[q2]<lowerlim)
    }
  if(q3>0)
    {
    for(j in 1:q3)
      {
      if(noprint==F)
        {
        cat(paste("variance of random",j,": ",round(theta1[q2+j],10),"\n",sep=""))
        }
      names1[q2+j]<-paste("random",j,sep="")
      }
    theta0[(q2+1):(q2+q3),1]<-theta1[(q2+1):(q2+q3)]
    theta0[(q2+1):(q2+q3),2]<-stopcrit[(q2+1):(q2+q3)]
    theta0[(q2+1):(q2+q3),3]<-1*(stopcrit[(q2+1):(q2+q3)]<lowerlim)
    }
  dimnames(theta0)<-list(names1,names2)
  if(outfile!="NA")
    {
    write.table(round(theta0,10),paste(outfile,"_variances.raw",sep=""),dimnames.write=T,sep=" ")
    }
  erg[[1]]<-round(theta0,10)
  names(erg)[1]<-"variances"
  }

#case 2: estimation with dispersion parameter
else
  {
#starting values for theta
  theta0<-rep(0,q2+q3+1)                                      
	# theta0[1]==phi, theta0[2:(q2+1)]==inverse smoothing parameter
	# theta0[(q2+2):(q2+q3+1)]==variances of random effects
  theta1<-c(var(dep-X%*%beta1[1:dim(X)[2]]),rep(startValue1,q2),rep(startValue2,q3))
#vector containing the signs of the current estimate of theta
#this vector is needed because of the unrestricted estimation of theta
  signs<-rep(1,q2+q3+1)

#score-function and expected fisher-information for theta
  s<-rep(0,q2+q3+1)
  Fi<-matrix(0,q2+q3+1,q2+q3+1)

#relative changes in beta / theta
  crit1<-1
  crit2<-1
  stopcrit<-rep(10,q2+q3+1)
  its<-rep(0,q2+q3+1)
  test<-T

#number of iteration
  k<-1

#linear predcitor
  if(family!="normal")
    {
    eta<-offset+X%*%beta1[1:dim(X)[2]]+B%*%beta1[(dim(X)[2]+1):(dim(X)[2]+dim(B)[2])]
    }

#start estimation loop
  while(test==T)
    {
    theta0<-theta1
    beta0<-beta1

#compute weights and y for working lmm
    if(family=="normal")
      {
      w1<-1/weight
      w<-rep(1/theta0[1],n)*weight
      }
    else if(family=="gamma")
      {
      mu<-exp(eta)
      w1<-1/weight
      w<-rep(1/theta0[1],n)*weight
      dinv<-exp(-eta)
      y<-as.matrix(eta-offset+dinv*(dep-mu))
      }
    else if(family=="poisson")
      {
      mu<-exp(eta)
      w1<-1/(mu*weight)
      w<-mu*weight*rep(1/theta0[1],n)
      dinv<-exp(-eta)
      y<-as.matrix(eta-offset+dinv*(dep-mu))
      }
    else if(family=="binomial")
      {
      if(link=="logit")
        {
        mu<-exp(eta)/(1+exp(eta))
        w1<-1/(mu*(1-mu)*weight)
        w<-mu*(1-mu)*weight*rep(1/theta0[1],n)
        dinv<-1/(mu*(1-mu))
        }
      else if(link=="probit")
        {
        mu<-pnorm(eta)
        w1<-1/(weight*dnorm(eta)^2/(mu*(1-mu)))
        w<-weight*dnorm(eta)^2/(mu*(1-mu))*rep(1/theta0[1],n)
        dinv<-1/dnorm(eta)
        }
      y<-as.matrix(eta-offset+dinv*(dep-mu))
      }

#fisher scoring step for beta
#if ranktest==T, H is tested for rank-deficiency and if H is singular, the diagonal elements are inflated (by multiplication with 1.0005)
    H<-rbind(cbind(t(X)%*%multdiagfront(X,w),t(X)%*%multdiagfront(B,w)),cbind(t(B)%*%multdiagfront(X,w),addtodiag(t(B)%*%multdiagfront(B,w),makeLambdainv(theta0[2:length(theta0)],bcut))))
    H1<-rbind(t(X)%*%(w*y),t(B)%*%(w*y))
    if(ranktest==T)
      {
      rank<-qr(H)$rank
      if(rank<dim(H)[1])
        {
        if(nowarnings==F)
          {
          cat("WARNING: fisher-information was found to be singular.\n")
          cat(paste("         rank is ",rank,", should be ",length(beta1),".\n",sep=""))
          cat(paste("         diagonal elements were multiplied with ",diagmult,".\n\n",sep=""))
          }
        H<-H+diag(diagmult*diag(H))
        }
      }
    beta1<-solve(H,H1)

#compute new linear predictor
    eta<-X%*%beta1[1:dim(X)[2],]+B%*%beta1[(dim(X)[2]+1):(dim(X)[2]+dim(B)[2]),]
    res<-as.matrix(y-eta)
    eta<-eta+offset

    if(method=="REML")
      {
#compute score-function and expected fisher-information for theta if estimation method is Restricted Maximum Lihkelihood

      Hinv<-solve(H)
      H<-addtodiag(H,-makeLambdainv(theta0[2:length(theta0)],bcut),dim(X)[2]+1,dim(H)[1])

#transform the restricted estimation problem to an unrestricted
      theta0<-signs*sqrt(theta0)
      w1<-w1*2*theta0[1]

      Hhelp1<-t(cbind(X,B))%*%multdiagfront(cbind(X,B),w^2*w1)
      Hhelp2<-t(cbind(X,B))%*%multdiagfront(cbind(X,B),w^3*w1^2)

      s[1]<--0.5*sum(w*w1)+0.5*spur2(Hhelp1,Hinv)+0.5*sum(w1*(res*w)^2)
      Fi[1,1]<-0.5*sum(w^2*w1^2)-spur2(Hhelp2,Hinv)+0.5*spur2(Hhelp1,Hinv%*%Hhelp1%*%Hinv)

      for(j in 1:(q2+q3))
        {
        Fi[1,(j+1)]<-0.5*spur(Hhelp1[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*theta0[j+1]*2)-spur(Hhelp1[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]%*%H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),]%*%Hinv*theta0[j+1]*2)+0.5*spur2(Hhelp1,Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]%*%H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),]%*%Hinv*theta0[j+1]*2)
        Fi[(j+1),1]<-Fi[1,(j+1)]

        s[j+1]<--0.5*spur(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*theta0[j+1]*2)+0.5*spur(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*theta0[j+1]*2)+sum((t(res*w)%*%B[,bcut[j]:(bcut[j+1]-1)])^2)*theta0[j+1]

        for(k in j:(q2+q3))
          {
          Fi[j+1,k+1]<-0.5*spur(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1)]%*%H[(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]*2*theta0[j+1]*2*theta0[k+1])-spur((H[(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)])%*%(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1)])*2*theta0[j+1]*2*theta0[k+1])+0.5*spur((H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1)])%*%(H[(dim(X)[2]+bcut[k]):(dim(X)[2]+bcut[k+1]-1),]%*%Hinv%*%H[,(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)])*2*theta0[j+1]*2*theta0[k+1])
          Fi[(k+1),(j+1)]<-Fi[(j+1),(k+1)]
          }
        }

      }
    else if(method=="ML")
      {
#compute score-function and expected fisher-information for theta if estimation method is Maximum Lihkelihood

      P<-solve(diag(as.vector(1/w))+multdiagback(B,makeLambda(theta1[2:length(theta1)],bcut))%*%t(B))

#transform the restricted estimation problem to an unrestricted
      theta0<-signs*sqrt(theta0)
      w1<-w1*2*theta0[1]

      s[1]<--0.5*spur(multdiagback(P,w1))+0.5*t(res*w*w1)%*%(w*res)
      Fi[1,1]<-0.5*spur(multdiagback(P,w1)%*%multdiagback(P,w1))

      for(j in 1:(q2+q3))
        {
        Fi[1,(j+1)]<-0.5*spur(P%*%(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j+1])%*%t(B[,bcut[j]:(bcut[j+1]-1)])%*%multdiagback(P,w1))
        Fi[(j+1),1]<-Fi[1,(j+1)]

        s[j+1]<--0.5*spur2(P,(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j+1])%*%t(B[,bcut[j]:(bcut[j+1]-1)]))+0.5*t(res*w)%*%(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j+1])%*%t(B[,bcut[j]:(bcut[j+1]-1)])%*%(w*res)

        for(k in j:(q2+q3))
          {
          Fi[(j+1),(k+1)]<-0.5*spur(P%*%(B[,bcut[j]:(bcut[j+1]-1)]*2*theta0[j+1])%*%t(B[,bcut[j]:(bcut[j+1]-1)])%*%P%*%(B[,bcut[k]:(bcut[k+1]-1)]*2*theta0[k+1])%*%t(B[,bcut[k]:(bcut[k+1]-1)]))
          Fi[(k+1),(j+1)]<-Fi[(j+1),(k+1)]
          }
        }
      }

#fisher-scoring step for theta
    theta1<-theta0+solve(Fi,s)
    signs<--1*(theta1<0)+1*(theta1>=0)

#transform theta back to original parametrisation
    theta1<-theta1^2
    theta0<-theta0^2

#stop estimation of theta[i] if stopcrit[i]<lowerlimit
    for(i in 1:(q2+q3))
      {
      stopcrit[i+1]<-sqrt(sum((B[,bcut[i]:(bcut[i+1]-1)]%*%beta1[(dim(X)[2]+bcut[i]):(dim(X)[2]+bcut[i+1]-1)])^2)/sum(eta^2))
      }
    theta1[stopcrit<lowerlim]<-theta0[stopcrit<lowerlim]
    its[stopcrit>lowerlim]<-k

#compute relative changes in beta and theta
    crit1<-sqrt(sum((beta0-beta1)^2)/sum(beta0^2))
    crit2<-sqrt(sum((theta0-theta1)^2)/sum(theta0^2))
    test<-(((crit1>eps)||(crit2>eps))&&(k<maxit))

    if(noprint==F)
      {
      cat(paste("iteration:",k,"\n"))
      cat(paste("relative changes in the regression coefficients:",round(crit1,7),"\n"))
      cat(paste("relative changes in the variance parameters:    ",round(crit2,7),"\n\n"))
      }
    k<-k+1
    }
  
  if(k>=maxit)
#algorithm did not converge
    {
    if(nowarnings==F)
      {
      cat(paste("WARNING: number of iterations reached ",maxit,".\n         algorithm did not converge.\n",sep=""))
      cat("         last two values of tau^2 were:\n")
      print(theta0[-1])
      print(theta1[-1])
      }
    }

  if(reg[1,1]!="NA")
    {
    q2<-q2-1
    }
  if(q2>0)
    {
    for(i in 1:q2)
      {
      if(stopcrit[1+i]<lowerlim && nowarnings==F)
        {
        cat(paste("WARNING: estimation of the variance of smooth",i," was stopped after iteration ",its[1+i],"\n         because its penalized part was small relative to the linear predictor.\n\n",sep=""))
        }
      }
    }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    if(stopcrit[1+q2]<lowerlim && nowarnings==F)
      {
      cat(paste("WARNING: estimation of the variance of the spatial effect was stopped after iteration ",its[q2+1],"\n         because its penalized part was small relative to the linear predictor.\n\n",sep=""))
      }
    }
  if(q3>0)
    {
    for(i in 1:q3)
      {
      if(stopcrit[1+q2+i]<lowerlim && nowarnings==F)
        {
        cat(paste("WARNING: estimation of the variance of random",i," was stopped after iteration ",its[q2+1+i],"\n         because its effect was small relative to the linear predictor.\n\n",sep=""))
        }
      }
    }

#extract fix and random effects    
  beta<-beta1[1:dim(X)[2]]
  alpha<-beta1[(dim(X)[2]+1):length(beta1)]


#store estimator of theta
  names2<-c("coeficient","stopcrit","stopped")
  if(noprint==F)
    {
    cat(paste("phi:",round(theta1[1],5),"\n"))
    }
  names1<-rep(" ",q2+q3+1)
  names1[1]<-"phi"
  theta0<-matrix(0,q2+q3+1,3)
  theta0[1:(q2+1),1]<-theta1[1:(q2+1)]
  theta0[1,2]<-NA
  theta0[1,3]<-NA
  if(reg[1,1]!="NA")
    {
    q2<-q2-1
    }
  if(q2>0)
    {
    for(j in 1:q2)
      {
      if(noprint==F)
        {
        cat(paste(paste("variance of smooth",j,sep=""),paste(":",round(theta1[1+j],10)),"\n",sep=""))
        }
      names1[1+j]<-paste("smooth",j,sep="")
      }
    theta0[2:(q2+1),2]<-stopcrit[2:(q2+1)]
    theta0[2:(q2+1),3]<-1*(stopcrit[2:(q2+1)]<lowerlim)
     }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    if(noprint==F)
      {
      cat(paste(paste("variance of the spatial effect:",round(theta1[1+q2],10)),"\n",sep=""))
      }
    names1[1+q2]<-"spatial"
    theta0[q2+1,2]<-stopcrit[q2+1]
    theta0[q2+1,3]<-1*(stopcrit[q2+1]<lowerlim)
    }
  if(q3>0)
    {
    for(j in 1:q3)
      {
      if(noprint==F)
        {
        cat(paste("variance of random",j,":",round(theta1[q2+j+1],10),"\n",sep=""))
        }
      names1[q2+j+1]<-paste("random",j,sep="")
      }
    theta0[(q2+2):(q2+q3+1),1]<-theta1[(q2+2):(q2+q3+1)]
    theta0[(q2+2):(q2+q3+1),2]<-stopcrit[(q2+2):(q2+q3+1)]
    theta0[(q2+2):(q2+q3+1),3]<-1*(stopcrit[(q2+2):(q2+q3+1)]<lowerlim)
    }
  dimnames(theta0)<-list(names1,names2)
  	# Schtzwerte und Konfidenzintervalle fr Varianz-/Glttungsparameter
  if(outfile!="NA")
    {
    write.table(round(theta0,10),paste(outfile,"_variances.raw",sep=""),dimnames.write=T,sep=" ")
    }
  erg[[1]]<-round(theta0,10)
  names(erg)[1]<-"variances"
}

#compute matrices that are needed to construct confidence intervals for the smooth, spatial and random effects
if((random[1,1]!="NA"||reg[1,1]!="NA")&&(ki==1))
  {
#compute confidence intervals from the marginal distribution of y
#help contains the diagonal elements of the covariance matrix of spatial and random effects
  if(reg[1,1]!="NA")
    {
    if(random[1,1]!="NA")
      {
      if(dispers==T)
        {
        help<-c(rep(theta0[(q2+1),1],length(regions)-1),rep(theta0[(2*q2+2):(2*q2+q3+1),1],each=length(randoms)))
        }
      else
        {
        help<-c(rep(theta0[q2,1],length(regions)-1),rep(theta0[(2*q2+1):(2*q2+q3),1],each=length(randoms)))
        }
      }
    else
      {
      if(dispers==T)
        {
        help<-rep(theta0[q2+1,1],length(regions)-1)
        }
      else
        {
        help<-rep(theta0[q2,1],length(regions)-1)
        }
      }
    q2<-q2-1
    q3<-q3+1
    }
  else
    {
    if(dispers==T)
      {
      help<-rep(theta0[(2*q2+2):(2*q2+q3+1),1],each=length(randoms))    
      }
    else
      {
      help<-rep(theta0[(2*q2+1):(2*q2+q3),1],each=length(randoms))    
      }
    }

#R is the marginal covariance matrix of y (as regards to spatial and random effects)
  Rinv<-solve(diag(as.vector(1/w))+B[,bcut[q2+1]:(bcut[q2+q3+1]-1)]%*%diag(help)%*%t(B[,bcut[q2+1]:(bcut[q2+q3+1]-1)]))
  if(dispers==F)
    {
    if(q2>0)
      {
      H<-rbind(cbind(t(X)%*%Rinv%*%X,t(X)%*%Rinv%*%B[,1:(bcut[q2+1]-1)]),cbind(t(B[,1:(bcut[q2+1]-1)])%*%Rinv%*%X,addtodiag(t(B[,1:(bcut[q2+1]-1)])%*%Rinv%*%B[,1:(bcut[q2+1]-1)],makeLambdainv(theta1[1:q2],bcut[1:(q2+1)]))))
      }
    else
      {
      H<-t(X)%*%Rinv%*%X
      }
    }
  else
    {
    if(q2>0)
      {
      H<-rbind(cbind(t(X)%*%Rinv%*%X,t(X)%*%Rinv%*%B[,1:(bcut[q2+1]-1)]),cbind(t(B[,1:(bcut[q2+1]-1)])%*%Rinv%*%X,addtodiag(t(B[,1:(bcut[q2+1]-1)])%*%Rinv%*%B[,1:(bcut[q2+1]-1)],makeLambdainv(theta1[2:(q2+1)],bcut[1:(q2+1)]))))
      }
    else
      {
      H<-t(X)%*%Rinv%*%X
      }
    }
  H1<-solve(H)
  if(q2>0)
    {
    H<-H1%*%t(cbind(X,B[,1:(bcut[q2+1]-1)]))%*%Rinv%*%cbind(X,B[,1:(bcut[q2+1]-1)])%*%H1
    }
  else
    {
    H<-H1
    }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    q3<-q3-1
    }
  }
else
#compute confidence intervals from the conditional distribution of y
  {
  if((family=="binomial"||family=="poisson")&&(dispers==F))
    {
    H<-rbind(cbind(t(X)%*%multdiagfront(X,w),t(X)%*%multdiagfront(B,w)),cbind(t(B)%*%multdiagfront(X,w),addtodiag(t(B)%*%multdiagfront(B,w),makeLambdainv(theta1,bcut))))
    }
  else
    {
    H<-rbind(cbind(t(X)%*%multdiagfront(X,w),t(X)%*%multdiagfront(B,w)),cbind(t(B)%*%multdiagfront(X,w),addtodiag(t(B)%*%multdiagfront(B,w),makeLambdainv(theta1[2:length(theta1)],bcut))))
    }
  H1<-solve(H)
  H<-H1%*%t(cbind(X,B))%*%multdiagfront(cbind(X,B),w)%*%H1
  }

#get smoothed effects and compute confidence intervals
if(smooth[1,1]!="NA")
  {
  if(reg[1,1]!="NA")
    {
    q2<-q2-1
    }
  names1<-rep(" ",6)
  h<-1
  for(j in 1:q2)
    {
    if(ord[j]==2)
      {
      smoothedEffects<-matrix(0,n,12)
      dimnames(smoothedEffects)<-list(NULL,c(paste("x",j,sep=""),paste("f",j,sep=""),cilstring,ciustring,"std","pvalue",ciblstring,cibustring,"stdb","pvalueb",paste("f",j,"lin",sep=""),paste("f",j,"minusf",j,"lin",sep="")))
      smoothedEffects[,2]<-X[,(q1+h)]*beta[q1+h]+B[,bcut[j]:(bcut[j+1]-1)]%*%alpha[bcut[j]:(bcut[j+1]-1)]
      Hhelp<-rbind(cbind(as.matrix(H[(q1+h),(q1+h)]),t(as.matrix(H[(q1+h),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]))),cbind(as.matrix(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(q1+h)]),as.matrix(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)])))
      e<-eigen(Hhelp,symmetric=T)
      e$values[e$values<0]<-0
      Hhelp<-e$vectors%*%diag(sqrt(e$values))%*%t(e$vectors)
      Hhelp1<-rbind(cbind(as.matrix(H1[(q1+h),(q1+h)]),t(as.matrix(H1[(q1+h),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)]))),cbind(as.matrix(H1[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(q1+h)]),as.matrix(H1[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)])))
      e<-eigen(Hhelp1,symmetric=T)
      e$values[e$values<0]<-0
      Hhelp1<-multdiagback(e$vectors,sqrt(e$values))%*%t(e$vectors)

      std<-sqrt(apply((cbind(X[,(q1+h)],B[,bcut[j]:(bcut[j+1]-1)])%*%Hhelp)^2,1,sum))
      std1<-sqrt(apply((cbind(X[,(q1+h)],B[,bcut[j]:(bcut[j+1]-1)])%*%Hhelp1)^2,1,sum))
      }
    else
      {
      smoothedEffects<-matrix(0,n,10)
      dimnames(smoothedEffects)<-list(NULL,c(paste("x",j,sep=""),paste("f",j,sep=""),cilstring,ciustring,"std","pvalue",ciblstring,cibustring,"stdb","pvalueb"))
      smoothedEffects[,2]<-B[,bcut[j]:(bcut[j+1]-1)]%*%alpha[bcut[j]:(bcut[j+1]-1)]
      e<-eigen(H[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)],symmetric=T)
      e$values[e$values<0]<-0
      Hhelp<-e$vectors%*%diag(sqrt(e$values))%*%t(e$vectors)
      e<-eigen(H1[(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1),(dim(X)[2]+bcut[j]):(dim(X)[2]+bcut[j+1]-1)],symmetric=T)
      e$values[e$values<0]<-0
      Hhelp1<-multdiagback(e$vectors,sqrt(e$values))%*%t(e$vectors)

      std<-sqrt(apply((B[,bcut[j]:(bcut[j+1]-1)]%*%Hhelp)^2,1,sum))
      std1<-sqrt(apply((B[,bcut[j]:(bcut[j+1]-1)]%*%Hhelp1)^2,1,sum))
      }

#arrange smoothedEffects according to the order of smooth[,j]
    smoothedEffects[,1]<-smooth[,j]
    o<-order(smoothedEffects[,1])
    smoothedEffects[,1]<-smoothedEffects[o,1]
    smoothedEffects[,2]<-smoothedEffects[o,2]
    std<-std[o]
    std1<-std1[o]

#get smoothedEffects according to distinct values of smooth[,j]
    h1<-c(Inf,smoothedEffects[1:(n-1),1])
    std<-std[h1!=smoothedEffects[,1]]
    std1<-std1[h1!=smoothedEffects[,1]]
    smoothedEffects<-smoothedEffects[h1!=smoothedEffects[,1],]

#center smoothedEffects and compute poitntwise confidence intervals and p-values
    beta[1]<-beta[1]+mean(smoothedEffects[,2])
    smoothedEffects[,2]<-center(smoothedEffects[,2])
    smoothedEffects[,3]<-smoothedEffects[,2]+qnorm(sig)*std
    smoothedEffects[,4]<-smoothedEffects[,2]-qnorm(sig)*std
    smoothedEffects[,5]<-std
    smoothedEffects[,6]<-pmin(pnorm(smoothedEffects[,2],mean=0,sd=std),1-pnorm(smoothedEffects[,2],mean=0,sd=std))
    smoothedEffects[,7]<-smoothedEffects[,2]+qnorm(sig)*std1
    smoothedEffects[,8]<-smoothedEffects[,2]-qnorm(sig)*std1
    smoothedEffects[,9]<-std1
    smoothedEffects[,10]<-pmin(pnorm(smoothedEffects[,2],mean=0,sd=std1),1-pnorm(smoothedEffects[,2],mean=0,sd=std1))


    if(ord[j]==2)
      {
#compute linear part of smoothedEffects and deviation from the linear part
      smoothedEffects[,11]<-center(smoothedEffects[,1]*beta[q1+h])
      smoothedEffects[,12]<-smoothedEffects[,2]-smoothedEffects[,11]
      h<-h+1
      }

#write estimates to the specified file
    if(outfile!="NA")
      {
      write.table(round(smoothedEffects,5),paste(outfile,"_smoothedEffects",j,".raw",sep=""),dimnames.write="colnames",sep=" ")
      }

#store results in erg
    erg[[1+j]]<-round(smoothedEffects,5)
    names(erg)[1+j]<-paste("smoothedEffects",j,sep="")

#plot the function
    if(plotf==T)
      {
      if(include.lin==T)
#plot with linear part of fj
        {
        plotsort1(smoothedEffects[,1],smoothedEffects[,2],smoothedEffects[,3],smoothedEffects[,4],smoothedEffects[,7],smoothedEffects[,8],smoothedEffects[,11],j)
        }
      else
#plot without linear part of fj
        {
        plotsort(smoothedEffects[,1],smoothedEffects[,2],smoothedEffects[,3],smoothedEffects[,4],smoothedEffects[,7],smoothedEffects[,8],j)
        }
      }
    }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    }
  }

#get spatial effects and compute confidence intervals
if(reg[1,1]!="NA")
  {
#design matrix for the (distinct) regions
  B1<-L%*%solve(t(L)%*%L)
  if(ki==1)
    {
#compute confidence intervals from the marginal distribution of y
    spatialEffects<-matrix(0,length(regions),2)
    spatialEffects[,1]<-regions
    spatialEffects[,2]<-B1%*%alpha[bcut[q2]:(bcut[q2+1]-1)]
    beta[1]<-beta[1]+mean(spatialEffects[,2])
    spatialEffects[,2]<-center(spatialEffects[,2])
    dimnames(spatialEffects)<-list(NULL,c("region","f"))
    }
  else
    {
#compute confidence intervals from the conditional distribution of y
    spatialEffects<-matrix(0,length(regions),10)
    spatialEffects[,1]<-regions
    spatialEffects[,2]<-B1%*%alpha[bcut[q2]:(bcut[q2+1]-1)]
    beta[1]<-beta[1]+mean(spatialEffects[,2])
    spatialEffects[,2]<-center(spatialEffects[,2])
    std<-sqrt(diag(B1%*%H[(dim(X)[2]+bcut[q2]):(dim(X)[2]+bcut[q2+1]-1),(dim(X)[2]+bcut[q2]):(dim(X)[2]+bcut[q2+1]-1)]%*%t(B1)))
    std1<-sqrt(diag(B1%*%H1[(dim(X)[2]+bcut[q2]):(dim(X)[2]+bcut[q2+1]-1),(dim(X)[2]+bcut[q2]):(dim(X)[2]+bcut[q2+1]-1)]%*%t(B1)))
    spatialEffects[,3]<-spatialEffects[,2]+qnorm(sig)*std
    spatialEffects[,4]<-spatialEffects[,2]-qnorm(sig)*std
    spatialEffects[,5]<-std
    spatialEffects[,6]<-pmin(pnorm(spatialEffects[,2],mean=0,sd=std),1-pnorm(spatialEffects[,2],mean=0,sd=std))
    spatialEffects[,7]<-spatialEffects[,2]+qnorm(sig)*std1
    spatialEffects[,8]<-spatialEffects[,2]-qnorm(sig)*std1
    spatialEffects[,9]<-std1
    spatialEffects[,10]<-pmin(pnorm(spatialEffects[,2],mean=0,sd=std1),1-pnorm(spatialEffects[,2],mean=0,sd=std1))
    dimnames(spatialEffects)<-list(NULL,c("region","f",cilstring,ciustring,"std","pvalue",ciblstring,cibustring,"stdb","pvalueb"))
    }

  if(plotmap==T)
#plot the spatial effect
    {
    drawmap(map=map,regionvar=spatialEffects[,1],plotvar=spatialEffects[,2],color=T,legend=T)
    }

#write estimates to the specified file
  if(outfile!="NA")
    {
    write.table(round(spatialEffects,5),paste(outfile,"_spatialEffects.raw",sep=""),dimnames.write="colnames",sep=" ")
    }

#store estimates in erg
  erg[[q2+1]]<-round(spatialEffects,5)
  names(erg)[q2+1]<-"spatialEffects"
  }

#get random effects and compute confidence intervals
if(random[1,1]!="NA")
  {
  if(ki==1)
#compute confidence intervals from the marginal distribution of y
    {
    randomEffects<-matrix(0,length(randoms),2)
    dimnames(randomEffects)<-list(NULL,c("id","f"))
    randomEffects[,1]<-randoms
    for(j in 1:q3)
      {
      randomEffects[,2]<-alpha[bcut[q2+j]:(bcut[q2+j+1]-1)]
      if(outfile!="NA")
        {
        write.table(round(randomEffects,5),paste(outfile,"_randomEffects",j,".raw",sep=""),dimnames.write="colnames",sep=" ")
        }
      erg[[q2+j+1]]<-round(randomEffects,5)
      names(erg)[q2+j+1]<-paste("randomEffects",j,sep="")
      }
    }
  else
    {
#compute confidence intervals from the conditional distribution of y
    randomEffects<-matrix(0,length(randoms),10)
    dimnames(randomEffects)<-list(NULL,c("region","f",cilstring,ciustring,"std","pvalue",ciblstring,cibustring,"stdb","pvalueb"))
    randomEffects[,1]<-randoms
    for(j in 1:q3)
      {
      randomEffects[,2]<-alpha[bcut[q2+j]:(bcut[q2+j+1]-1)]
      std<-sqrt(diag(H[(dim(X)[2]+bcut[q2+j]):(dim(X)[2]+bcut[q2+j+1]-1),(dim(X)[2]+bcut[q2+j]):(dim(X)[2]+bcut[q2+j+1]-1)]))
      std1<-sqrt(diag(H1[(dim(X)[2]+bcut[q2+j]):(dim(X)[2]+bcut[q2+j+1]-1),(dim(X)[2]+bcut[q2+j]):(dim(X)[2]+bcut[q2+j+1]-1)]))
      randomEffects[,3]<-randomEffects[,2]+qnorm(sig)*std
      randomEffects[,4]<-randomEffects[,2]-qnorm(sig)*std
      randomEffects[,5]<-std
      randomEffects[,6]<-pmin(pnorm(randomEffects[,2],0,sd=std),1-pnorm(randomEffects[,2],0,sd=std))
      randomEffects[,7]<-randomEffects[,2]+qnorm(sig)*std1
      randomEffects[,8]<-randomEffects[,2]-qnorm(sig)*std1
      randomEffects[,9]<-std1
      randomEffects[,10]<-pmin(pnorm(randomEffects[,2],0,sd=std1),1-pnorm(randomEffects[,2],0,sd=std1))
      if(outfile!="NA")
        {
        write.table(round(randomEffects,5),paste(outfile,"_randomEffects",j,".raw",sep=""),dimnames.write="colnames",sep=" ")
        }
      erg[[q2+j+1]]<-round(randomEffects,5)
      names(erg)[q2+j+1]<-paste("randomEffects",j,sep="")
      }
    }
  }

#compute variances of the fixed effects
if(q1>1)
  {
  std<-sqrt(diag(H[1:q1,1:q1]))
  std1<-sqrt(diag(H1[1:q1,1:q1]))
  }
else
  {
  std<-sqrt(H[1,1])
  std1<-sqrt(H1[1,1])
  }

#get fixed effects and compute confidence intervals
fixedEffect<-as.matrix(beta[1:q1])
cil<-fixedEffect+qnorm(sig)*std
cir<-fixedEffect-qnorm(sig)*std
cil1<-fixedEffect+qnorm(sig)*std1
cir1<-fixedEffect-qnorm(sig)*std1
names1<-rep(" ",q1)
for(j in 1:q1)
  {
  if(noprint==F)
    {
    cat(paste(paste("beta",(j-1),sep=""),paste(":",round(beta[j],5)),"\n",sep=""))
    }
  names1[j]<-paste("beta",(j-1),sep="")
  }
names2<-c("coefficient",cilstring,ciustring,"std","pvalue",ciblstring,cibustring,"stdb","pvalueb")
fixedEffect<-cbind(fixedEffect,cil,cir,std,pmin(pnorm(fixedEffect[,1],mean=0,sd=std),1-pnorm(fixedEffect[,1],mean=0,sd=std)),cil1,cir1,std1,pmin(pnorm(fixedEffect[,1],mean=0,sd=std1),1-pnorm(fixedEffect[,1],mean=0,sd=std1)))

#store parameter corresponding to the linear effect of fj
h<-1
if(reg[1,1]!="NA")
  {
  q2<-q2-1
  }
if(q2>0)
  {
  for(j in 1:q2)
    {
    if(ord[j]==2)
      {
      fixedEffect<-rbind(fixedEffect,c(beta[q1+h],NA,NA,NA,NA,NA,NA,NA,NA))
      h<-h+1
      names1<-c(names1,paste("fbeta",j,sep=""))
      }
    }
  }
if(reg[1,1]!="NA")
  {
  q2<-q2+1
  }

dimnames(fixedEffect)<-list(names1,names2)
if(outfile!="NA")
  {
  write.table(round(fixedEffect,5),paste(outfile,"_fixedEffects.raw",sep=""),dimnames.write=T,sep=" ")
  }
erg[[q2+q3+2]]<-round(fixedEffect,5)
names(erg)[q2+q3+2]<-"fixedEffects"


#compute the log-likelihood 
if((family=="normal") && (log.like==T || lrtest==T))
  {
  log.like<-matrix(0,1,1)
  beta<-beta1[1:dim(X)[2]]

#residuals and marginal covariance matrix of y (as regards to smooth, spatial and random effects)
  res<-y-X%*%beta
  V<-diag(as.vector(1/w))+multdiagback(B,makeLambda(theta1[2:length(theta1)],bcut))%*%t(B)
  Vinv<-solve(V)
  log.like[1,1]<--0.5*(logdet(V)+t(res)%*%Vinv%*%res)
  if(method=="REML")
    {
    log.like[1,1]<-log.like[1,1]-0.5*logdet(t(X)%*%Vinv%*%X)
    }
#  dimnames(log.like)<-list(NULL,c("ML","REML"))

  if(noprint==F)
    {
    if(method=="ML")
      {
      cat(paste("log-likelihood: ",round(log.like[1,1],5),"\n",sep=""))
      }
    else if(method=="REML")
      {
      cat(paste("restricted log-likelihood: ",round(log.like[1,1],5),"\n",sep=""))
      }
    }
  if(outfile!="NA")
    {
    write.table(round(log.like,5),paste(outfile,"_log.like.raw",sep=""),dimnames.write=F,sep=" ")
    }
  erg[[q2+q3+3]]<-round(log.like,5)
  names(erg)[q2+q3+3]<-"log.like"
  q2<-q2+1
  }


#perform likelihood-ratio tests on the different model components based on the (incorrect) theory of self/liang (1987)
if(lrtest==T)
{
q2<-q2-1
if(method=="ML")
  {
  if(reg[1,1]=="NA")
    {
    testmat<-matrix(0,sum(ord)+q3,2)
    names2<-rep(" ",sum(ord)+q3)
    }
  else
    {
    testmat<-matrix(0,sum(ord)+q3+1,2)
    names2<-rep(" ",sum(ord)+q3+1)
    }
  }
else
  {
  if(reg[1,1]=="NA")
    {
    testmat<-matrix(0,q2+q3,2)
    names2<-rep(" ",q2+q3)
    }
  else
    {
    testmat<-matrix(0,q2+q3+1,2)
    names2<-rep(" ",q2+q3+1)
    }
  } 
h<-1
if(smooth[1,1]!="NA")
  {
  if(reg[1,1]!="NA")
    {
    q2<-q2-1
    }
  if(q2>1)
    {
    for(j in 1:q2)
      {
      if(ord[j]==2)
        {
        if(noprint==F)
          {
          cat(paste("\nre-estimating the model with influence of x",j," assumed to be linear\n",sep=""))
          }
        if(dim(fix)[2]==1)
          {
          test<-ggamm(dep=dep,family=family,link=link,fix=smooth[,j],smooth=smooth[,-j],nknot=nknot[-j],ord=ord[-j],deg=deg[-j],reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
          }
        else
          {
          test<-ggamm(dep=dep,family=family,link=link,fix=cbind(fix[,-1],smooth[,j]),smooth=smooth[,-j],nknot=nknot[-j],ord=ord[-j],deg=deg[-j],reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
          }
        test<-2*(log.like-test)
        testmat[h,1]<-test
        testmat[h,2]<-chi.mix(arg=testmat[h,1])
        if(noprint==F)
          {
          if(method=="ML")
            {
            cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
            cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
            }
          else if(method=="REML")
            {
            cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
            cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
            }
          }
        names2[h]<-paste("x",j," linear",sep="")
        h<-h+1
        }
      }
    for(j in 1:q2)
      {
      if(ord[j]==2 && method=="REML")
        {
        }
      else
        {
        if(noprint==F)
          {
          cat(paste("\nre-estimating the model without influence of x",j,"\n",sep=""))
          }
        if(dim(fix)[2]==1)
          {
          test<-ggamm(dep=dep,family=family,link=link,smooth=smooth[,-j],nknot=nknot[-j],ord=ord[-j],deg=deg[-j],reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
          }
        else
          {
          test<-ggamm(dep=dep,family=family,link=link,fix=fix[,-1],smooth=smooth[,-j],nknot=nknot[-j],ord=ord[-j],deg=deg[-j],reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
          }
        test<-2*(log.like-test)
        testmat[h,1]<-test
        if(ord[j]==2)
          {
          testmat[h,2]<-chi.mix(arg=testmat[h,1],df=c(1,2))
          if(noprint==F)
            {
            cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
            cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
            }
          }
        else
          {
          testmat[h,2]<-chi.mix(arg=testmat[h,1])
          if(noprint==F)
            {
            cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
            cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
            }
          }
        names2[h]<-paste("x",j," absent",sep="")
        h<-h+1
        }
      }
    }
  else
    {
    if(ord==2)
      {
      if(noprint==F)
        {
        cat("\nre-estimating the model with influence of x1 assumed to be linear\n")
        }
      if(dim(fix)[2]==1)
        {
        test<-ggamm(dep=dep,family=family,link=link,fix=smooth,reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
        }
      else
        {
        test<-ggamm(dep=dep,family=family,link=link,fix=cbind(fix[,-1],smooth),reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
        }
      test<-2*(log.like-test)
      testmat[h,1]<-test
      testmat[h,2]<-chi.mix(arg=testmat[h,1])
      if(noprint==F)
        {
        if(method=="ML")
          {
          cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
          cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
          }
        else if(method=="REML")
          {
          cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
          cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
          }
        }
      names2[h]<-"x1 linear"
      h<-h+1
      }
    if(ord==2 && method=="REML")
      {
      }
    else
      {
      if(noprint==F)
        {
        cat("\nre-estimating the model without influence of x1\n")
        }
      if(dim(fix)[2]==1)
        {
        test<-ggamm(dep=dep,family=family,link=link,reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
        }
      else
        {
        test<-ggamm(dep=dep,family=family,link=link,fix=fix[,-1],reg=reg,regions=regions,pmatrix=pmatrix,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
        }
      test<-2*(log.like-test)
      testmat[h,1]<-test[1,1]
      if(ord==2)
        {
        testmat[h,2]<-chi.mix(arg=testmat[h,1],df=c(1,2))
        if(noprint==F)
          {
          cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
          cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
          }
        }
      else
        {
        testmat[h,2]<-chi.mix(arg=testmat[h,1])
        if(noprint==F)
          {
          cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
          cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
          }
        }
      names2[h]<-"x1 absent"
      h<-h+1
      }
    }
  if(reg[1,1]!="NA")
    {
    q2<-q2+1
    }
  }

if(reg[1,1]!="NA")
  {
  if(noprint==F)
    {
    cat("\nre-estimating the model without spatial effect\n")
    }
  test<-ggamm(dep=dep,family=family,link=link,fix=fix,smooth=smooth,nknot=nknot,ord=ord,deg=deg,random=random,id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
  test<-2*(log.like-test)
  testmat[h,1]<-test
  testmat[h,2]<-chi.mix(arg=testmat[h,1])
  if(noprint==F)
    {
    if(method=="ML")
      {
      cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
      cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
      }
    else if(method=="REML")
      {
      cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
      cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
      }
    }
  names2[h]<-"spat absent"
  h<-h+1
  }
if(random[1,1]!="NA")
  {
  if(q3>1)
    {
    for(j in 1:q3)
      {
      if(noprint==F)
        {
        cat(paste("\nre-estimating the model without influence of z",j,"\n",sep=""))
        }
      test<-ggamm(dep=dep,family=family,link=link,fix=fix,smooth=smooth,nknot=nknot,ord=ord,deg=deg,reg=reg,regions=regions,pmatrix=pmatrix,random=random[,-j],id=id,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
      test<-2*(log.like-test)
      testmat[h,1]<-test
      testmat[h,2]<-chi.mix(arg=testmat[h,1])
      if(noprint==F)
        {
        if(method=="ML")
          {
          cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
          cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
          }
        else if(method=="REML")
          {
          cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
          cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
          }
        }
      names2[h]<-paste("z",j," absent",sep="")
      h<-h+1
      }
    }
  else
    {
    if(noprint==F)
      {
      cat("\nre-estimating the model without influence of z1\n")
      }
    test<-ggamm(dep=dep,family=family,link=link,fix=fix,smooth=smooth,nknot=nknot,ord=ord,deg=deg,reg=reg,regions=regions,pmatrix=pmatrix,weight=weight,offset=offset,eps=eps,startValue1=startValue1,startValue2=startValue2,maxit=maxit,ranktest=ranktest,dispers=dispers,log.like=T,noprint=T,method=method,nowarnings=T)$log.like
    test<-2*(log.like-test)
    testmat[h,1]<-test
    testmat[h,2]<-chi.mix(arg=testmat[h,1])
    if(noprint==F)
      {
      if(method=="ML")
        {
        cat(paste("likelihood ratio: ",testmat[h,1],"\n",sep=""))
        cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
        }
      else if(method=="REML")
        {
        cat(paste("restricted likelihood ratio: ",testmat[h,1],"\n",sep=""))
        cat(paste("p-value: ",testmat[h,2],"\n",sep=""))
        }
      }
    names2[h]<-"z1 absent"
    h<-h+1
    }
  }
if(method=="ML")
  {
  dimnames(testmat)<-list(names2,c("lr","p-values"))
  }
else
  {
  dimnames(testmat)<-list(names2,c("rlr","p-values"))
  }
erg[[q2+q3+4]]<-testmat
names(erg)[q2+q3+4]<-"lr-tests"
if(outfile!="NA")
  {
  write.table(lrtest,paste(outfile,"_lr-tests.raw",sep=""),dimnames.write="colnames",sep=" ")
  }
q2<-q2+2
}

pred<-matrix(0,n,2)
dimnames(pred)<-list(NULL,c("eta","mu"))

if(family=="normal")
  {
  mu<-eta
  }
else if(family=="gamma")
  {
  mu<-exp(eta)
  }
else if(family=="poisson")
  {
  mu<-exp(eta)
  }
else if(family=="binomial")
  {
  if(link=="logit")
    {
    mu<-exp(eta)/(1+exp(eta))
    }
  else if(link=="probit")
    {
    mu<-pnorm(eta)
    }
  }
pred[,1]<-as.matrix(eta)
pred[,2]<-as.matrix(mu)
if(outfile!="NA")
  {
  write.table(round(pred,5),paste(outfile,"_predict.raw",sep=""),dimnames.write="colnames",sep=" ")
  }

erg[[q2+q3+3]]<-round(pred,5)
names(erg)[q2+q3+3]<-"predict"

erg[[q2+q3+4]]<-k-1
names(erg)[q2+q3+4]<-"iterations"
if(outfile!="NA")
  {
  write.table(as.matrix(erg[[q2+q3+4]]),paste(outfile,"_iterations.raw",sep=""),dimnames.write="colnames",sep=" ")
  }

return(erg)
}



